@@ -29,8 +29,7 @@ void RestitutionTest::Initialize()
 		BodyCreationSettings settings(sphere, RVec3(-50.0f + i * 10.0f, 20.0f, -20.0f), Quat::sIdentity(), EMotionType::Dynamic, Layers::MOVING);
 		settings.mRestitution = 0.1f * i;
 		settings.mLinearDamping = 0.0f;
-		Body &body = *mBodyInterface->CreateBody(settings);
-		mBodyInterface->AddBody(body.GetID(), EActivation::Activate);
+		mBodyInterface->CreateAndAddBody(settings, EActivation::Activate);
 	}
 
 	for (int i = 0; i <= 10; ++i)
@@ -38,7 +37,6 @@ void RestitutionTest::Initialize()
 		BodyCreationSettings settings(box, RVec3(-50.0f + i * 10.0f, 20.0f, 20.0f), Quat::sIdentity(), EMotionType::Dynamic, Layers::MOVING);
 		settings.mRestitution = 0.1f * i;
 		settings.mLinearDamping = 0.0f;
-		Body &body = *mBodyInterface->CreateBody(settings);
-		mBodyInterface->AddBody(body.GetID(), EActivation::Activate);
+		mBodyInterface->CreateAndAddBody(settings, EActivation::Activate);
 	}
 }
