@@ -15,16 +15,21 @@ public:
 
 	// See: Test
 	virtual void			Initialize() override;
+	virtual void			ProcessInput(const ProcessInputParams &inParams) override;
 	virtual void			PrePhysicsUpdate(const PreUpdateParams &inParams) override;
+	virtual void			SaveInputState(StateRecorder &inStream) const override;
+	virtual void			RestoreInputState(StateRecorder &inStream) override;
 
 	virtual void			GetInitialCamera(CameraState &ioState) const override;
-	virtual RMat44			GetCameraPivot(float inCameraHeading, float inCameraPitch) const override;
+	virtual RMat44			GetCameraPivot(float inCameraHeading, float inCameraPitch) const override { return mCameraPivot; }
 
 private:
 	static constexpr float	cMaxSteeringAngle = DegreesToRadians(30);
 
 	using EAxis = SixDOFConstraintSettings::EAxis;
 
+	void					UpdateCameraPivot();
+
 	enum class EWheel : int
 	{
 		LeftFront,
@@ -39,4 +44,9 @@ private:
 
 	Body *					mCarBody;
 	Ref<SixDOFConstraint>	mWheels[int(EWheel::Num)];
+	RMat44					mCameraPivot = RMat44::sIdentity();	///< The camera pivot, recorded before the physics update to align with the drawn world
+
+	// Player input
+	float					mSteeringAngle = 0.0f;
+	float					mSpeed = 0.0f;
 };
