@@ -6,6 +6,8 @@
 #include <Jolt/Core/FPFlushDenormals.h>
 #include <atomic>
 
+#ifndef JPH_CPU_WASM
+
 // Implemented as a global atomic so the compiler can't optimize it to a constant
 extern atomic<float> TestFltMin;
 atomic<float> TestFltMin = FLT_MIN;
@@ -38,3 +40,5 @@ TEST_SUITE("FlushDenormalsTests")
 		TestFltMin = 1.0f;
 	}
 }
+
+#endif // JPH_CPU_WASM
